@@ -34,7 +34,14 @@ def get_provided_kallisto_path() -> Optional[str]:
     Returns:
         Path to the binary, `None` if not found
     """
-    bin_filename = 'kallisto.exe' if PLATFORM == 'windows' else 'kallisto'
+    bin_name = 'kallisto'
+    if '_KALLISTO_OPTOFF' in globals():
+        if _KALLISTO_OPTOFF:
+            bin_name = f'{bin_name}_optoff'
+    if '_KALLISTO_KMER_64' in globals():
+        if _KALLISTO_KMER_64:
+            bin_name = f'{bin_name}_k64'
+    bin_filename = f'{bin_name}.exe' if PLATFORM == 'windows' else bin_name
     path = os.path.join(BINS_DIR, PLATFORM, CPU, 'kallisto', bin_filename)
     if not os.path.isfile(path):
         return None
@@ -54,11 +61,18 @@ def get_provided_bustools_path() -> Optional[str]:
     return path
 
 
+def set_special_kallisto_binary(k64: bool, optoff: bool):
+    global _KALLISTO_KMER_64
+    global _KALLISTO_OPTOFF
+    _KALLISTO_KMER_64 = k64
+    _KALLISTO_OPTOFF = optoff
+
+
 def get_compiled_kallisto_path(alias: str = COMPILED_DIR) -> Optional[str]:
     """Finds platform-dependent kallisto binary compiled with `compile`.
 
     Args:
-        Alias: Alias of compiled binary.
+        alias: Alias of compiled binary.
 
     Returns:
         Path to the binary, `None` if not found
@@ -74,7 +88,7 @@ def get_compiled_bustools_path(alias: str = COMPILED_DIR) -> Optional[str]:
     """Finds platform-dependent bustools binary compiled with `compile`.
 
     Args:
-        Alias: Alias of compiled binary.
+        alias: Alias of compiled binary.
 
     Returns:
         Path to the binary, `None` if not found
