@@ -201,6 +201,7 @@ def run_executable(
         if not quiet and p.returncode != returncode:
             logger.error('\n'.join(out))
             raise sp.CalledProcessError(p.returncode, ' '.join(command))
+        # logger.info(stdout)
 
     return (p, stdout, stderr) if wait else p
 
@@ -412,6 +413,83 @@ def read_t2g(t2g_path: str) -> Dict[str, Tuple[str, ...]]:
     return t2g
 
 
+def obtain_gene_names(
+    t2g_path: str,
+    gene_names_list: Union[str, List[str]],
+    verbose: Optional[bool] = True,
+    clean_dups: Optional[bool] = True
+) -> List[str]:
+    """Given a transcript-to-gene mapping path and list of gene IDs,
+    return a list of cleaned-up gene names (wherein blank names are simply
+    replaced by the corresponding gene ID, as are duplicate names if specified)
+
+    Args:
+        t2g_path: Path to t2g
+        gene_names_list: List of gene IDs or path to list of gene IDs
+        verbose: Whether to warn about the number of blank names, defaults to `True`
+        clean_dups: Whether to convert duplicate names to gene IDs, defaults to `True`
+
+    Returns:
+        List of gene names
+    """
+    is_geneid_path = isinstance(gene_names_list, str)
+    var_names = []
+    if is_geneid_path:
+        if not os.path.exists(gene_names_list):
+            return []
+        with open_as_text(gene_names_list, 'r') as f:
+            var_names = [line.strip() for line in f]
+    else:
+        var_names = gene_names_list
+
+    t2g = read_t2g(t2g_path)
+    id_to_name = {}
+    for transcript, attributes in t2g.items():
+        if len(attributes) > 1:
+            id_to_name[attributes[0]] = attributes[1]
+    # Locate duplicates:
+    names_set = set([])
+    duplicates_set = set([])
+    if clean_dups:
+        for gene_id in var_names:
+            if id_to_name.get(gene_id):
+                if id_to_name[gene_id] in names_set:
+                    duplicates_set.add(id_to_name[gene_id])
+                names_set.add(id_to_name[gene_id])
+    # Now make list of cleaned-up gene names:
+    gene_names = []
+    n_no_name = 0
+    for gene_id in var_names:
+        if id_to_name.get(gene_id) and not (id_to_name[gene_id]
+                                            in duplicates_set):
+            gene_names.append(id_to_name[gene_id])
+        else:  # blank names and duplicate names are considered missing
+            gene_names.append(gene_id)
+            n_no_name += 1
+    if n_no_name > 0 and verbose:
+        logger.warning(
+            f'{n_no_name} gene IDs do not have corresponding valid gene names. '
+            'These genes will use their gene IDs instead.'
+        )
+    return gene_names
+
+
+def write_list_to_file(strings: List[str], str_path: str) -> str:
+    """Write out a list of strings.
+
+    Args:
+        strings: List of strings to output
+        str_path: Path to output
+
+    Returns:
+        Path to written file
+    """
+    with open_as_text(str_path, 'w') as out:
+        for s in strings:
+            out.write(f'{s}\n')
+    return str_path
+
+
 def collapse_anndata(
     adata: anndata.AnnData, by: Optional[str] = None
 ) -> anndata.AnnData:
@@ -477,7 +555,10 @@ def import_tcc_matrix_as_anndata(
     barcodes_path: str,
     ec_path: str,
     txnames_path: str,
-    threads: int = 8
+    threads: int = 8,
+    loom: bool = False,
+    loom_names: List[str] = None,
+    batch_barcodes_path: Optional[str] = None,
 ) -> anndata.AnnData:
     """Import a TCC matrix as an Anndata object.
 
@@ -486,13 +567,25 @@ def import_tcc_matrix_as_anndata(
         barcodes_path: Path to the barcodes txt file
         genes_path: Path to the ec txt file
         txnames_path: Path to transcripts.txt generated by `kallisto bus`
+        threads: Number of threads, defaults to `8`
+        loom: Whether to prepare anndata for loom file, defaults to `False`
+        loom_names: Names for cols and rows in anndata, defaults to `None`
+        batch_barcodes_path: Path to barcodes prefixed with sample ID,
+            defaults to `None`
 
     Returns:
         A new Anndata object
     """
+    name_column = 'transcript_ids' if not loom else loom_names[1]
+    bc_name = 'barcode' if not loom else loom_names[0]
     df_barcodes = pd.read_csv(
-        barcodes_path, index_col=0, header=None, names=['barcode']
+        barcodes_path, index_col=0, header=None, names=[bc_name]
     )
+    if (batch_barcodes_path):
+        df_batch_barcodes = pd.read_csv(
+            batch_barcodes_path, index_col=0, header=None, names=[bc_name]
+        )
+        df_barcodes.index = df_batch_barcodes.index + df_barcodes.index
     df_ec = pd.read_csv(
         ec_path,
         index_col=0,
@@ -522,7 +615,7 @@ def import_tcc_matrix_as_anndata(
     transcript_ids = []
     for future in futures:
         transcript_ids += future.result()
-    df_ec['transcript_ids'] = pd.Categorical(transcript_ids)
+    df_ec[name_column] = pd.Categorical(transcript_ids)
     df_ec.drop('transcripts', axis=1, inplace=True)
     return anndata.AnnData(
         X=scipy.io.mmread(matrix_path).tocsr(), obs=df_barcodes, var=df_ec
@@ -536,6 +629,9 @@ def import_matrix_as_anndata(
     t2g_path: Optional[str] = None,
     name: str = 'gene',
     by_name: bool = False,
+    loom: bool = False,
+    loom_names: List[str] = None,
+    batch_barcodes_path: Optional[str] = None,
 ) -> anndata.AnnData:
     """Import a matrix as an Anndata object.
 
@@ -549,15 +645,26 @@ def import_matrix_as_anndata(
         name: Name of the columns, defaults to "gene"
         by_name: Aggregate counts by name instead of ID. `t2g_path` must be
             provided and contain names.
+        loom: Whether to prepare anndata for loom file, defaults to `False`
+        loom_names: Names for cols and rows in anndata, defaults to `None`
+        batch_barcodes_path: Path to barcodes prefixed with sample ID,
+            defaults to `None`
 
     Returns:
         A new Anndata object
     """
+    name_column = f'{name}_id' if not loom else loom_names[1]
+    bc_name = 'barcode' if not loom else loom_names[0]
     df_barcodes = pd.read_csv(
-        barcodes_path, index_col=0, header=None, names=['barcode']
+        barcodes_path, index_col=0, header=None, names=[bc_name]
     )
+    if (batch_barcodes_path):
+        df_batch_barcodes = pd.read_csv(
+            batch_barcodes_path, index_col=0, header=None, names=[bc_name]
+        )
+        df_barcodes.index = df_batch_barcodes.index + df_barcodes.index
     df_genes = pd.read_csv(
-        genes_path, header=None, index_col=0, names=[f'{name}_id'], sep='\t'
+        genes_path, header=None, index_col=0, names=[name_column], sep='\t'
     )
     df_genes.index = df_genes.index.astype(
         str
@@ -567,29 +674,12 @@ def import_matrix_as_anndata(
         anndata.AnnData(X=mtx.tocsr(), obs=df_barcodes, var=df_genes)
     )
 
-    name_column = f'{name}_name'
-    n_no_name = 0
-    if t2g_path:
-        t2g = read_t2g(t2g_path)
-        id_to_name = {}
-        for transcript, attributes in t2g.items():
-            if len(attributes) > 1:
-                id_to_name[attributes[0]] = attributes[1]
-        gene_names = []
-        for gene_id in adata.var_names:
-            if id_to_name.get(gene_id):  # blank names are considered missing
-                gene_names.append(id_to_name[gene_id])
-            else:
-                gene_names.append(gene_id)
-                n_no_name += 1
+    if t2g_path and by_name:
+        gene_names = obtain_gene_names(
+            t2g_path, adata.var_names.to_list(), False
+        )
         adata.var[name_column] = pd.Categorical(gene_names)
 
-        if n_no_name > 0:
-            logger.warning(
-                f'{n_no_name} gene IDs do not have corresponding gene names. '
-                'These genes will use their gene IDs instead.'
-            )
-
     return (
         collapse_anndata(adata, by=name_column)
         if name_column in adata.var.columns and by_name else adata
@@ -597,7 +687,9 @@ def import_matrix_as_anndata(
 
 
 def overlay_anndatas(
-    adata_spliced: anndata.AnnData, adata_unspliced: anndata.AnnData
+    adata_spliced: anndata.AnnData,
+    adata_unspliced: anndata.AnnData,
+    adata_ambiguous: anndata.AnnData = None
 ) -> anndata.AnnData:
     """'Overlays' anndata objects by taking the intersection of the obs and var
     of each anndata.
@@ -610,6 +702,7 @@ def overlay_anndatas(
     Args:
         adata_spliced: An Anndata object
         adata_unspliced: An Anndata object
+        adata_ambiguous: An Anndata object, default `None`
 
     Returns:
         A new Anndata object
@@ -618,18 +711,20 @@ def overlay_anndatas(
     var_idx = adata_spliced.var.index.intersection(adata_unspliced.var.index)
     spliced_intersection = adata_spliced[obs_idx][:, var_idx]
     unspliced_intersection = adata_unspliced[obs_idx][:, var_idx]
+    a_layers = {
+        'mature': spliced_intersection.X,
+        'nascent': unspliced_intersection.X
+    }
+    sum_X = spliced_intersection.X + unspliced_intersection.X
+    ambiguous_intersection = None
+    if adata_ambiguous is not None:
+        ambiguous_intersection = adata_ambiguous[obs_idx][:, var_idx]
+        a_layers.update({'ambiguous': ambiguous_intersection.X})
+        sum_X = sum_X + ambiguous_intersection.X
 
     df_obs = unspliced_intersection.obs
     df_var = unspliced_intersection.var
-    return anndata.AnnData(
-        X=spliced_intersection.X,
-        layers={
-            'spliced': spliced_intersection.X,
-            'unspliced': unspliced_intersection.X
-        },
-        obs=df_obs,
-        var=df_var
-    )
+    return anndata.AnnData(X=sum_X, layers=a_layers, obs=df_obs, var=df_var)
 
 
 def sum_anndatas(
@@ -664,6 +759,154 @@ def sum_anndatas(
     )
 
 
+def do_sum_matrices(
+    mtx1_path, mtx2_path, out_path, mm=False, header_line=None
+) -> str:
+    """Sums up two matrices given two matrix files.
+
+    Args:
+        mtx1_path: First matrix file path
+        mtx2_path: Second matrix file path
+        out_path: Output file path
+        mm: Whether to allow multimapping (i.e. decimals)
+        header_line: The header line if we have it
+
+    Returns:
+        Output file path
+    """
+    logger.info('Summing matrices into {}'.format(out_path))
+    n = 0
+    header = []
+    with open_as_text(mtx1_path,
+                      'r') as f1, open_as_text(mtx2_path,
+                                               'r') as f2, open(out_path,
+                                                                'w') as out:
+        eof1 = eof2 = pause1 = pause2 = False
+        nums = [0, 0, 0]
+        nums1 = nums2 = to_write = None
+        if header_line:
+            out.write("%%MatrixMarket matrix coordinate real general\n%\n")
+        while not eof1 or not eof2:
+            s1 = f1.readline() if not eof1 and not pause1 else '%'
+            s2 = f2.readline() if not eof2 and not pause2 else '%'
+            if not s1:
+                pause1 = eof1 = True
+            if not s2:
+                pause2 = eof2 = True
+            _nums1 = _nums2 = []
+            if not eof1 and s1[0] != '%':
+                _nums1 = s1.split()
+                if not mm:
+                    _nums1[0] = int(_nums1[0])
+                    _nums1[1] = int(_nums1[1])
+                    _nums1[2] = int(float(_nums1[2]))
+                else:
+                    _nums1[0] = int(_nums1[0])
+                    _nums1[1] = int(_nums1[1])
+                    _nums1[2] = float(_nums1[2])
+            if not eof2 and s2[0] != '%':
+                _nums2 = s2.split()
+                if not mm:
+                    _nums2[0] = int(_nums2[0])
+                    _nums2[1] = int(_nums2[1])
+                    _nums2[2] = int(float(_nums2[2]))
+                else:
+                    _nums2[0] = int(_nums2[0])
+                    _nums2[1] = int(_nums2[1])
+                    _nums2[2] = float(_nums2[2])
+            if nums1 is not None:
+                _nums1 = nums1
+                nums1 = None
+            if nums2 is not None:
+                _nums2 = nums2
+                nums2 = None
+            if eof1 and eof2:
+                # Both mtxs are done
+                break
+            elif eof1:
+                # mtx1 is done
+                nums = _nums2
+                pause2 = False
+            elif eof2:
+                # mtx2 is done
+                nums = _nums1
+                pause1 = False
+            elif eof1 and eof2:
+                # Both mtxs are done
+                break
+            # elif (len(_nums1) != len(_nums2)):
+            #    # We have a problem
+            #    raise Exception("Summing up two matrix files failed")
+            elif not _nums1 or not _nums2:
+                # We have something other than a matrix line
+                continue
+            elif not header:
+                # We are at the header line and need to read it in
+                if (_nums1[0] != _nums2[0] or _nums1[1] != _nums2[1]):
+                    raise Exception(
+                        "Summing up two matrix files failed: Headers incompatible"
+                    )
+                else:
+                    header = [_nums1[0], _nums1[1]]
+                if header_line:
+                    out.write(header_line)
+                continue
+            elif (_nums1[0] > _nums2[0]
+                  or (_nums1[0] == _nums2[0] and _nums1[1] > _nums2[1])):
+                # If we're further in mtx1 than mtx2
+                nums = _nums2
+                pause1 = True
+                pause2 = False
+                nums1 = _nums1
+                nums2 = None
+            elif (_nums2[0] > _nums1[0]
+                  or (_nums2[0] == _nums1[0] and _nums2[1] > _nums1[1])):
+                # If we're further in mtx2 than mtx1
+                nums = _nums1
+                pause2 = True
+                pause1 = False
+                nums2 = _nums2
+                nums1 = None
+            elif _nums1[0] == _nums2[0] and _nums1[1] == _nums2[1]:
+                # If we're at the same location in mtx1 and mtx2
+                nums = _nums1
+                nums[2] += _nums2[2]
+                pause1 = pause2 = False
+                nums1 = nums2 = None
+            else:
+                # Shouldn't happen
+                raise Exception(
+                    "Summing up two matrix files failed: Assertion failed"
+                )
+            # Write out a line
+            _nums_prev = to_write
+            if (_nums_prev and _nums_prev[0] == nums[0]
+                    and _nums_prev[1] == nums[1]):
+                nums[2] += _nums_prev[2]
+                pause1 = pause2 = False
+                to_write = [nums[0], nums[1], nums[2]]
+            else:
+                if to_write:
+                    if header_line:
+                        if mm and to_write[2].is_integer():
+                            to_write[2] = int(to_write[2])
+                        out.write(
+                            f'{to_write[0]} {to_write[1]} {to_write[2]}\n'
+                        )
+                    n += 1
+                to_write = [nums[0], nums[1], nums[2]]
+        if to_write:
+            if header_line:
+                if mm and to_write[2].is_integer():
+                    to_write[2] = int(to_write[2])
+                out.write(f'{to_write[0]} {to_write[1]} {to_write[2]}\n')
+            n += 1
+    if not header_line:
+        header_line = f'{header[0]} {header[1]} {n}\n'
+        do_sum_matrices(mtx1_path, mtx2_path, out_path, mm, header_line)
+    return out_path
+
+
 def restore_cwd(func: Callable) -> Callable:
     """Function decorator to decorate functions that change the current working
     directory. When such a function is decorated with this function, the
